<?php

namespace App\Http\Controllers;

use DB;
use Log;
use App\Users;
use Illuminate\Http\Request;
use App\Http\ResponseHelper;
use App\Jobs\GenerateUserTokenJob;
use Osiset\BasicShopifyAPI\BasicShopifyAPI;
use Osiset\BasicShopifyAPI\Options;
use Osiset\BasicShopifyAPI\Session;
use Config;

class CustomerCreationWebhookController extends Controller
{

    private $shopify_api_version;
    private $shopifyWebhookSecret;

    public function __construct() {

        $this->helper = new ResponseHelper();
        $this->shopify_api_version = env('SHOPIFY_API_VERSION', true);
        $this->shopifyWebhookSecret = env('SHOPIFY_WEBHOOK_SECRET', true);
    }

    /**
     * Shopify customer auth token creation
     */
    public function processWebhook() {

        $hmac_header = $_SERVER['HTTP_X_SHOPIFY_HMAC_SHA256'];
        $requestData = file_get_contents('php://input');
        $verified = $this->verify_shopify_webhook($requestData, $hmac_header);
        $webhookStatus = var_export($verified, true);

        if ($webhookStatus) {

            $receivedData = json_decode($requestData);

            if(empty($receivedData->id)) {

                Log::channel('generateUserTokenLog')->info('Shopify customer create 
                webhook response is not correct, WebhookResponse:'.$requestData);
            } else {

                $randomString = $this->generateRandomString(16);

                $lastInsertedId = DB::table('webhooks')->insertGetId(
                    [
                        'webhookContent' => $requestData,
                        'authToken' => $randomString
                    ]
                );
                $queueResponse = dispatch(new GenerateUserTokenJob($requestData, $randomString));

                ignore_user_abort(true);
                header("Content-Length: 0");
                header("Connection: Close");
                flush();
            }
        } else {

            Log::channel('generateUserTokenLog')->info('Shopify customer create webhook validation 
            failed, WebhookResponse:'.$requestData);
        }
    }

    /**
     * Shopify webhook verification
     */
    public function verify_shopify_webhook($data, $hmac_header) {

        $calculated_hmac = base64_encode(hash_hmac('sha256', $data, $this->shopifyWebhookSecret, true));
        return hash_equals($hmac_header, $calculated_hmac);
    }

    /**
     * Generate random string
     */
    public function generateRandomString($length) {

        $characters = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
        $charactersLength = strlen($characters);
        $randomString = '';
        for ($i = 0; $i < $length; $i++) {
            $randomString .= $characters[rand(0, $charactersLength - 1)];
        }
        return $randomString;
    }

    /**
     * Generate and create customer metafield with auth token cron
     */
    public function handleWebhookData() {

        Log::channel('cron')->info('handleWebhookData webhook');
        
        $password   = Config::get('shopify.passwordWebhook');
        $shopDomain = Config::get('shopify.shopDomainWebhook');

        $options = new Options();
        $options->setVersion($this->shopify_api_version);

        // Create the client and session
        $api = new BasicShopifyAPI($options);
        $api->setSession(new Session($shopDomain, $password));


        $webhookDetails = DB::table('webhooks')->where([
            ['status', '=', '0']
        ])->get();

        if(!empty($webhookDetails)){
            $updateIds = array();
            foreach($webhookDetails as $webhookData){
                $webhookContent = $webhookData->webhookContent;
                $authToken  = $webhookData->authToken;
                $webhookId = $webhookData->id;
                $webhook = json_decode($webhookContent);

                $email = $webhook->email;
                $customerId  = $webhook->id;

                $metafieldArray = array();

                $temp['key']        = 'authToken';
                $temp['value']      = $authToken;
                $temp['type']       = 'single_line_text_field';
                $temp['namespace']  = 'global';
                
                array_push($metafieldArray, $temp);

                $user = DB::table('users')->where('user_email', '=', $email)->get();
        
                if(!isset($user[0]->user_email)){
                    // create user in DB 
                    $userId = DB::table('users')->insertGetId([
                        'user_email' => $email,
                        'status' => 1,
                        'auth_token' => $authToken
                    ]);
                } else {
                    $userId = DB::table('users')->where('user_email', $email)->update(['auth_token' => $authToken]);  
                }

                $method = 'GET';
                $url    = '/admin/api/'.$this->shopify_api_version.'/customers/'.$customerId.'/metafields.json';
                $params = null;
                $result = $api->rest($method, $url, $params);

                if(empty($result->errors))
                {
                    $metafields = $result->body->metafields;

                    if(!empty($metafields)){
                        $updated = false;
                        foreach($metafields as $metafield){
            
                            $metaId        = $metafield->id;
                            $metaKey       = $metafield->key;
                            $metaNameSpace = $metafield->namespace;
            
                            if($metaKey == 'authToken' && $metaNameSpace == 'global'){
            
                                // update Metafield
                                $updateRequest = [];
                                $updateRequest['metafield']['id']         = $metaId;
                                $updateRequest['metafield']['value']      = $authToken;
                                $updateRequest['metafield']['type']       = "single_line_text_field";
            
                                $params = $updateRequest;
                        
                                $method = 'PUT';
                                $url    = '/admin/api/'.$this->shopify_api_version.'/metafields/'.$metaId.'.json';
                                $result = $api->rest($method, $url, $params); 
            
                                if(!empty($result->body->metafield)){
                                    // Updated successfully
                                    $updated = true;
                                    Log::channel('cron')->info("Updated metafield successfully for the user ".$customerId." with value ".$authToken);
                                    $updateIds[] = $webhookId;   
                                }
                                else{
                                    // Not updated
                                    $updated = false;
                                    Log::channel('cron')->info("Failed to update metafield for the user ".$customerId." with value ".$authToken);
                                    Log::channel('cron')->info(json_encode($result));                                    
                                }
                        
                            }
            
                        }
            
                        if($updated == false){
            
                            $createMetaResult = $this->createCustomerMetafield($metafieldArray, $customerId, $authToken, $api, $webhookId);
                            if($createMetaResult){
                                $updateIds[] = $webhookId;
                            }
                        }
                    }else{
                        
                        $createMetaResult = $this->createCustomerMetafield($metafieldArray, $customerId, $authToken, $api, $webhookId);
                        if($createMetaResult){
                            $updateIds[] = $webhookId;
                        }
                
                    }

                }

            }
            if(!empty($updateIds)){
                DB::table('webhooks')->whereIn('id',$updateIds)->update(['status' => 1]);
            }
        }
        else{
            Log::channel('cron')->info("Empty Webhook Data");
        }
        
    }

    function createCustomerMetafield($metafield, $customerId, $authToken, $api, $webhookId){
        
        $params = array(
            'customer' => array(
                'id' => $customerId,
                'metafields' => $metafield
            )
        );

        $method = 'PUT';
        $url = '/admin/api/'.$this->shopify_api_version.'/customers/'.$customerId.'.json';

        $result = $api->rest($method, $url, $params);

        if(!empty($result->body->customer)){
            // Updated customer successfully by adding new metafield
            $updated = true;
            Log::channel('cron')->info("Updated customer successfully by adding new metafield ".$customerId." with value ".$authToken);
            return true;
        } else {
            // Not updated
            $updated = false;
            Log::channel('cron')->info("Failed to update customer metafield for the user ".$customerId." with value ".$authToken);
            Log::channel('cron')->info(json_encode($result));
            return false;
        }
    }

}
