<?php

namespace App\Http\Controllers;

use Log;
use Carbon\Carbon;
use App\Models\RechargeApi;
use App\Models\Shopify\Order;
use App\Helper\Helper;
use Illuminate\Http\Request;

class SubscriptionsAndOrderHandlingV3Controller extends Controller {

    private $rechargeObj;
    private $orderObj;
    private $logDirectory;
    private $logPath;
    private $logFiles;

    public function __construct() {

        $this->rechargeObj = new RechargeApi('getAllSubscriptions');
        $this->orderObj = new Order('getAllSubscriptions');
    }

    /**
     * Get All subscriptions and order history details
     */
    public function getAllSubscriptionsv3(Request $request) {

        $requestData = $request->input('data');
        if (empty($requestData)) {

            $returnData = [
                'error' => 'Invalid endpoint request'
            ];            
            Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Invalid endpoint request");
            return $this->returnData(600, "error", $returnData);
        }
        $receivedData = json_decode($requestData);
        
        $subscriptionId = !empty($receivedData->subscriptionId) ? $receivedData->subscriptionId : '';
        $emailAddress = !empty($receivedData->email) ? trim($receivedData->email) : '';
        $requestedHash = !empty($receivedData->hash_id) ? trim($receivedData->hash_id) : '';

        if (empty($subscriptionId) && empty($emailAddress) && empty($requestedHash)) {

            $returnData = [
                'error' => 'Invalid endpoint request'
            ];            
            Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Invalid endpoint request, 
            endpoint request:".json_encode($receivedData));
            return $this->returnData(600, "error", $returnData);
        }

        if ((empty($subscriptionId) && !empty($emailAddress)) || (!empty($subscriptionId) && empty($emailAddress))) {

            $returnData = [
                'error' => 'Subscription Id or the email address is missing in the endpoint request'
            ];            
            Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Subscription Id or 
            the email address is missing in the endpoint request, endpoint request:".json_encode($receivedData));
            return $this->returnData(600, "error", $returnData);
        }
        
        $regularSubscriptionAndOrderDetails = [];
        $giftSubscriptionAndOrders = [];
        $sortedSubscriptionsAndOrders = [];
        // Get regular subscription and order details
        if (!empty($requestedHash)) {
            $regularSubscriptionAndOrderDetails = $this->getRegularSubscriptionAndOrderDetailsV3($requestedHash);
            if ($regularSubscriptionAndOrderDetails['status'] == 'failed') {

                Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: ".
                $regularSubscriptionAndOrderDetails['logInfo']);
                return $this->returnData(600, "error", $regularSubscriptionAndOrderDetails['response']);
            }
        }

        if ((!empty($subscriptionId) || isset($subscriptionId)) && !empty($emailAddress)) {

            $giftSubscriptionAndOrders = $this->getGiftSubscriptionAndOrderDetailsV3($subscriptionId, $emailAddress);
        }

        if ((empty($regularSubscriptionAndOrderDetails) && empty($giftSubscriptionAndOrders)) || 
        (empty($regularSubscriptionAndOrderDetails['subscriptions']) && empty($giftSubscriptionAndOrders['subscriptions']))) {

            $returnData = [
                'error' => 'Not found any regular or gift subscriptions'
            ];            
            Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Not found any regular 
            or gift subscriptions, endpoint request:".json_encode($receivedData));
            return $this->returnData(600, "error", $returnData);
        }

        // Sort subscriptions and orders in descending order
        $sortedSubscriptionsAndOrders = $this->sortSubscriptionsAndOrders($regularSubscriptionAndOrderDetails, 
        $giftSubscriptionAndOrders);

        $returnData = [
            'message' => 'Successfully get the subscriptions and order details',
            'data' => $sortedSubscriptionsAndOrders
        ];

        Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: ".json_encode($returnData));
        return $this->returnData(200, "success", $returnData);
    }

    /**
     * Get regular subscriptions and order details
     */
    public function getRegularSubscriptionAndOrderDetailsV3($requestedHash) {

        // Get subscriptions
        $returnDataArray = [];
        // Get customer details by hash
        $customerDetails = $this->rechargeObj->getRechargeCustomerDetailsByHash($requestedHash);
        if (!empty($customerDetails->error) || !empty($customerDetails->errors) ||
            empty($customerDetails->customers)) {

            $rechargeError  = "Unknown error";
            if (!empty($customerDetails->error)) {

                $rechargeError = json_encode($customerDetails->error);
            } else if (!empty($customerDetails->errors)) {

                $rechargeError = json_encode($customerDetails->errors);
            } else if (empty($customerDetails->customers)) {

                $rechargeError = 'Failed to get the customer details from recharge by using the provided hash_id';
            }
            $returnData = [
                'error' => $rechargeError
            ];

            $returnDataArray['status'] = "failed";
            $returnDataArray['logInfo'] = "getAllSubscriptionsv3: Failed to get the customer details from 
            the recharge(rechargeHash:".$requestedHash."), Error: (".$rechargeError.")";
            $returnDataArray['response'] = $returnData;
            return $returnDataArray;
        }

        $regularSubscriptionDetails = [];
        $rechargeCustomerIds = [];
        foreach ($customerDetails->customers as $customerKey => $customerValue) {

            $rechargeCustomerId = !empty($customerValue->id) ? trim($customerValue->id) : '';
            $rechargeCustomerIds[] = $rechargeCustomerId;

            // Get all the subscriptions by customerId
            $subscriptionDetails = $this->rechargeObj->getAllSubscriptionsByCustomerIDV2($rechargeCustomerId);
            if (!empty($subscriptionDetails->error) || !empty($subscriptionDetails->errors) ||
                empty($subscriptionDetails->subscriptions)) {

                $rechargeError  = "Unknown error";
                if (!empty($subscriptionDetails->error)) {

                    $rechargeError = json_encode($subscriptionDetails->error);
                } else if (!empty($subscriptionDetails->errors)) {

                    $rechargeError = json_encode($subscriptionDetails->errors);
                } else if (empty($subscriptionDetails->subscriptions)) {

                    $rechargeError = 'Failed to get the subscriptions from recharge by customerId';
                }
                $returnData = [
                    'error' => $rechargeError
                ];

                Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Failed to get 
                the subscriptions from recharge by customerId(customerId:".$rechargeCustomerId."), Error: (".
                $rechargeError.")");
                continue;
            }

            $regularSubscriptionDetails[] = $subscriptionDetails->subscriptions;
            //sleep(1);
        }
        if (!empty($regularSubscriptionDetails)) {

            foreach ($regularSubscriptionDetails as $subKey => $subValue) {

                foreach ($subValue as $key => $value) {

                    $isPrepaidStatus = !empty($value->is_prepaid) ? trim($value->is_prepaid) : '';
                    //$checkSkuIsPrepaid = !empty($value->sku) ? trim(strtolower($value->sku)) : '';
                    if ($isPrepaidStatus == 1) {

                        $value->gift = "true";
                    } else {

                        $value->gift = "false";
                    }
                }
                $returnDataArray['subscriptions'] = $subValue;
            }
        }

        if (empty($returnDataArray['subscriptions']) || !isset($returnDataArray['subscriptions'])) {

            $returnDataArray['status'] = "success";
            return $returnDataArray;
        }

        // Get subscription success orders
        $subscriptionOrders = [];
        $subscriptionOrders = $this->getRechargeOrderDetails($rechargeCustomerIds);
        if (empty($subscriptionOrders['orders']) || !isset($subscriptionOrders['orders'])) {

            $returnDataArray['status'] = "success";
            return $returnDataArray;
        }
        $subscriptionOrders['orders'] = json_decode(json_encode($subscriptionOrders['orders']), FALSE);
        foreach ($subscriptionOrders['orders'] as $orderKey => $orderValue) {

            $lipProperties = !empty($orderValue->line_items[0]->properties) ? $orderValue->line_items[0]->properties: '';

            if (!empty($lipProperties)) {

                foreach ($lipProperties as $propKey => $propValue) {

                    $propName = !empty($propValue->name) ? trim($propValue->name) : '';
                    $propValue = !empty($propValue->value) ? trim($propValue->value) : '';

                    if ($propName == 'giftee_email' && !empty($propValue)) {

                        $orderValue->gift = "true";
                        break;
                    } else {

                        $orderValue->gift = "false";
                    }
                }
            }
        }
        $returnDataArray['orders'] = $subscriptionOrders['orders'];
        $returnDataArray['status'] = "success";
        return $returnDataArray;
    }

    /**
     * Get gift subscriptions and order details
     */
    public function getGiftSubscriptionAndOrderDetailsV3($subscriptionId, $gifteeEmail) {
        
        $returnDataArray = [];
        $giftSubscriptionDetails = [];
        $giftSubscriptionOrderDetails = [];
        // Get subscriptions by ID
        foreach ($subscriptionId as $subKey => $subValue) {

            $giftSubscriptions = $this->rechargeObj->getSubscriptionV2($subValue);
            if (!empty($giftSubscriptions->error) || !empty($giftSubscriptions->errors) ||
                empty($giftSubscriptions)) {

                $rechargeError  = "Unknown error";
                if (!empty($giftSubscriptions->error)) {

                    $rechargeError = json_encode($giftSubscriptions->error);
                } else if (!empty($giftSubscriptions->errors)) {

                    $rechargeError = json_encode($giftSubscriptions->errors);
                } else if (empty($giftSubscriptions)) {

                    $rechargeError = 'Failed to get the subscription details by subscriptionId';
                }
                $returnData = [
                    'error' => $rechargeError
                ];

                $returnDataArray['status'] = "failed";
                $returnDataArray['logInfo'] = "getAllSubscriptionsv3: Failed to get the subscription 
                details by subscriptionId(subscriptionId:".$subValue."), Error: (".$rechargeError.")";
                $returnDataArray['response'] = $returnData;
                return $returnDataArray;
            }

            if (!empty($giftSubscriptions->subscription->properties) || isset($giftSubscriptions->subscription->properties)) {

                foreach ($giftSubscriptions->subscription->properties as $giftPropKey => $giftPropValue) {

                    $propertyName = !empty($giftPropValue->name) ? trim(strtolower($giftPropValue->name)) : '';
                    if ($propertyName == 'giftee_email') {

                        $propertyValue = !empty($giftPropValue->value) ? trim(strtolower($giftPropValue->value)) : '';
                        if ($propertyValue == trim(strtolower($gifteeEmail))) {
                            
                            $giftSubscriptions->subscription->gift = "true";
                            $giftSubscriptionDetails[] = $giftSubscriptions->subscription;
                        }
                    }
                }
            }
            //sleep(1);
        }

        if (empty($giftSubscriptionDetails) || !isset($giftSubscriptionDetails)) {

            $returnDataArray['status'] = "success";
            return $returnDataArray;
        }

        // Get gift subscription orders
        foreach ($giftSubscriptionDetails as $giftKey => $giftValue) {

            $subscriptionAddressId = !empty($giftValue->address_id) ? $giftValue->address_id : '';
            // Get all order details from recharge by address ID
            $allOrderDetails = $this->rechargeObj->getRechargeOrdersByAddressIdV2($subscriptionAddressId);
            if (!empty($allOrderDetails->error) || !empty($allOrderDetails->errors) ||
                empty($allOrderDetails) || empty($allOrderDetails->orders)) {

                $rechargeError  = "Unknown error";
                if (!empty($allOrderDetails->error)) {

                    $rechargeError = json_encode($allOrderDetails->error);
                } else if (!empty($allOrderDetails->errors)) {

                    $rechargeError = json_encode($allOrderDetails->errors);
                } else if (empty($allOrderDetails) || empty($allOrderDetails->orders)) {

                    $rechargeError = 'Failed to get the order details from recharge';
                }

                $returnData = [
                    'error' => $rechargeError
                ];

                Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Failed to get the order details from 
                the recharge(subscriptionId:".$giftValue->id.", addressId:".$subscriptionAddressId."),
                Error: (".$rechargeError.")");
            }

            $upcomingQueuedOrderScheduledDate = '';
            $upcomingQueuedOrderShippingAddress = '';
            $rechargeOrdersShopifyOrderIds = [];
            foreach ($allOrderDetails->orders as $orderKey => $orderValue) {

                $orderStatus = !empty($orderValue->status) ? trim(strtolower($orderValue->status)) : '';
                if ($orderStatus == 'success') {

                    $orderValue->gift = "true";
                    $rechargeOrdersShopifyOrderIds[] = $orderValue->shopify_order_id;
                    $giftSubscriptionOrderDetails[] = $orderValue;
                } else if ($orderStatus == 'queued') {

                    $upcomingQueuedOrderScheduledDate = $orderValue->scheduled_at;
                    $upcomingQueuedOrderShippingAddress = $orderValue->shipping_address;
                }
            }
            $giftValue->scheduled_at = $upcomingQueuedOrderScheduledDate;
            $giftValue->shipping_address = $upcomingQueuedOrderShippingAddress;
            //sleep(1);
        }

        // Get and add Shopify order extracted status to the order
        $shopifyOrderDetails = [];
        $shopifyOrderDetails = $this->getShopifyOrderDetails($rechargeOrdersShopifyOrderIds);
        if (!empty($shopifyOrderDetails) || isset($shopifyOrderDetails)) {

            foreach ($giftSubscriptionOrderDetails as $orderKey => $orderValue) {

                $shopifyOrderId = '';
                $shopifyOrderId = $orderValue->shopify_order_id;

                foreach ($shopifyOrderDetails as $orderTagKey => $orderTagValue) {

                    if (trim($shopifyOrderId) == trim($orderTagKey)) {

                        $giftSubscriptionOrderDetails[$orderKey]->extracted = $orderTagValue;
                    }
                }
            }
        }

        $returnDataArray['status'] = "success";
        $returnDataArray['subscriptions'] = $giftSubscriptionDetails;
        $returnDataArray['orders'] = $giftSubscriptionOrderDetails;
        return $returnDataArray;
    }

    /**
     * Sort regular and gift subscriptions and orders
     */
    public function sortSubscriptionsAndOrders($regularSubscriptionsAndOrders, $giftSubscriptionsAndOrders) {

        $combinedSubscriptions = [];
        $combinedOrders = [];
        $subscriptionsWithCreateDate = [];
        $ordersWithCreatedDate = [];
        $subscriptionsWithStrTime = [];
        $ordersWithStrTime = [];
        $allSubscriptionDetails = [];
        $allOrderHistoryDetails = [];
        $subscriptionsAndOrdersCombined = [];

        if (!empty($regularSubscriptionsAndOrders['subscriptions'])) {
            foreach ($regularSubscriptionsAndOrders['subscriptions'] as $regSubKey => $regSubValue) {
                $combinedSubscriptions[$regSubValue->id] = $regSubValue;
                $subscriptionsWithCreateDate[$regSubValue->id] = $regSubValue->created_at;
            }
        }

        if (!empty($giftSubscriptionsAndOrders['subscriptions'])) {
            foreach ($giftSubscriptionsAndOrders['subscriptions'] as $giftSubKey => $giftSubValue) {
                $combinedSubscriptions[$giftSubValue->id] = $giftSubValue;
                $subscriptionsWithCreateDate[$giftSubValue->id] = $giftSubValue->created_at;
            }
        }

        if (!empty($regularSubscriptionsAndOrders['orders'])) {
            foreach ($regularSubscriptionsAndOrders['orders'] as $regOrderKey => $regOrderValue) {
                $combinedOrders[$regOrderValue->id] = $regOrderValue;
                $ordersWithCreatedDate[$regOrderValue->id] = $regOrderValue->created_at;
            }
        }

        if (!empty($giftSubscriptionsAndOrders['orders'])) {
            foreach ($giftSubscriptionsAndOrders['orders'] as $giftOrderKey => $giftOrderValue) {
                $combinedOrders[$giftOrderValue->id] = $giftOrderValue;
                $ordersWithCreatedDate[$giftOrderValue->id] = $giftOrderValue->created_at;
            }
        }

        foreach ($subscriptionsWithCreateDate as $createSubKey => $createSubValue) {

            $dateStrToTime = '';
            $dateStrToTime = strtotime($createSubValue);
            $subscriptionsWithStrTime[$createSubKey] = $dateStrToTime;
        }
        arsort($subscriptionsWithStrTime);

        foreach ($ordersWithCreatedDate as $createOrderKey => $createOrderValue) {

            $dateStrToTime = '';
            $dateStrToTime = strtotime($createOrderValue);
            $ordersWithStrTime[$createOrderKey] = $dateStrToTime;
        }
        arsort($ordersWithStrTime);

        foreach ($subscriptionsWithStrTime as $subStrKey => $subStrValue) {

            $allSubscriptionDetails[] = $combinedSubscriptions[$subStrKey];
        }

        foreach ($ordersWithStrTime as $orderStrKey => $orderStrValue) {

            $allOrderHistoryDetails[] = $combinedOrders[$orderStrKey];
        }

        $subscriptionsAndOrdersCombined['subscriptions'] = $allSubscriptionDetails;
        $subscriptionsAndOrdersCombined['orders'] = $allOrderHistoryDetails;
        return $subscriptionsAndOrdersCombined;
    }

    /**
     * Get recharge orders by customer Id
     */
    public function getRechargeOrderDetails($rechargeCustomerIds) {

        Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Started to read the orders");
        $rechargeOrderDetails = [];
        $orderDetails = [];
        $pageNo = 1;
        $nextPageLink = '';
        $limit  = 200;
        foreach ($rechargeCustomerIds as $customerKey => $customerId) {
            do {
                try{
                    //sleep(1);
                    Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: Read recharge 
                    order records page -". $pageNo);
                    $rechargeOrders = $this->rechargeObj->getOrdersByCustomerId($customerId, $limit, $nextPageLink);
                    //sleep(1);
                }
                catch (Exception $e){
                    $errorText = "Error form recharge when fetching the order details and error is - ".$e->getMessage();
                    Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: ".$errorText);
                }
                $rechargeOrdersShopifyOrderIds = [];
                if (!empty($rechargeOrders) && isset($rechargeOrders)) {

                    foreach ($rechargeOrders['orders'] as $order){

                        if (!empty($order['id'])) {

                            $rechargeOrdersShopifyOrderIds[] = $order['shopify_order_id'];
                            $orderDetails[] = $order;
                        }
                    }
                    
                    // Get and add Shopify order extracted status to the order
                    $shopifyOrderDetails = [];
                    $shopifyOrderDetails = $this->getShopifyOrderDetails($rechargeOrdersShopifyOrderIds);
                    if (!empty($shopifyOrderDetails) || isset($shopifyOrderDetails)) {

                        foreach ($orderDetails as $orderKey => $orderValue) {

                            $shopifyOrderId = '';
                            $shopifyOrderId = $orderValue['shopify_order_id'];

                            foreach ($shopifyOrderDetails as $orderTagKey => $orderTagValue) {

                                if (trim($shopifyOrderId) == trim($orderTagKey)) {

                                    $orderDetails[$orderKey]['extracted'] = $orderTagValue;
                                }
                            }
                        }
                    }                

                    $nextPageLink = !empty($rechargeOrders['next']) ? $rechargeOrders['next'] : "";
                    if (empty($rechargeOrders['next'])) {
                        break;
                    }
                    $orderCount = !empty($rechargeOrders['orders']) ? 
                    count($rechargeOrders['orders']) : 0;
                }
                $pageNo++;

            } while($orderCount == $limit);
        }

        $rechargeOrderDetails['orders'] = $orderDetails;

        return $rechargeOrderDetails;
    }

    /**
     * Get Shopify order by order ids
     */
    public function getShopifyOrderDetails($rechargeOrdersShopifyOrderIds) {

        $shopifyExtractedTagStatus = [];
        $incrementCount = 0;
        $shopifyIds = '';
        $numberOfIdsToPass = (int)50;
        $totalNumberOfShopifyOrderIds = (int)count($rechargeOrdersShopifyOrderIds);
        $loopExecutionCount = round($totalNumberOfShopifyOrderIds / $numberOfIdsToPass);
        $loopExecutionCount = ($loopExecutionCount == 0) ? 1 : $loopExecutionCount;
        for ($i = 0; $i < $loopExecutionCount; $i++) { 

            $shopifyIdsToGetDetails = [];
            $loopConditionLimit = 0;

            if ($incrementCount == 0) {
                $loopConditionLimit = $numberOfIdsToPass;
            } else {
                $loopConditionLimit = (int)($incrementCount + $numberOfIdsToPass);
            }

            for ($j = $incrementCount; $j < $loopConditionLimit; $j++) { 

                if (empty($rechargeOrdersShopifyOrderIds[$j])) {
                    break;
                }
                $shopifyIdsToGetDetails[] = $rechargeOrdersShopifyOrderIds[$j];
                $incrementCount++;
            }

            $shopifyIds = implode(",", $shopifyIdsToGetDetails);
            $shopifyOrderDetails = $this->orderObj->getOrderDetailsByIds($shopifyIds);
            if (!empty($shopifyOrderDetails['error']) || !empty($shopifyOrderDetails['errors']) ||
                empty($shopifyOrderDetails) || empty($shopifyOrderDetails['orders'])) {

                $shopifyError  = "Unknown error";
                if (!empty($shopifyOrderDetails['error'])) {

                    $shopifyError = json_encode($shopifyOrderDetails['error']);
                } else if (!empty($shopifyOrderDetails['errors'])) {

                    $shopifyError = json_encode($shopifyOrderDetails['errors']);
                } else if (empty($shopifyOrderDetails) || empty($shopifyOrderDetails['orders'])) {

                    $shopifyError = 'Failed to get the order details from Shopify';
                }
                $returnData = [
                    'error' => $shopifyError
                ];

                Log::channel('subscriptionsAndOrderHandlingLog')->info("getAllSubscriptionsv3: 
                Failed to get the order details from Shopify(OrderIds:".$shopifyIds."), Error: 
                (".$shopifyError.")");
                return $this->returnData(600, "error", $returnData);
            }

            if (!empty($shopifyOrderDetails['orders']) || isset($shopifyOrderDetails['orders'])) {

                foreach ($shopifyOrderDetails['orders'] as $shOrderKey => $shOrderValue) {

                    $orderTags = '';
                    $orderExtractedStatus = '';
                    $orderTags = !empty($shOrderValue['tags']) ? $shOrderValue['tags'] : '';
                    $orderExtractedStatus = $this->checkOrderTagWithExtracted($orderTags);
                    if ($orderExtractedStatus) {

                        $shopifyExtractedTagStatus[$shOrderValue['id']] = "true";
                    } else {

                        $shopifyExtractedTagStatus[$shOrderValue['id']] = "false";
                    }
                }
            }
            $shopifyOrderDetails = [];
            //sleep(1);
        }

        return $shopifyExtractedTagStatus;
    }

    /**
     * Check Shopify order tags
     */
    public function checkOrderTagWithExtracted($orderTags) {

        $explodeTags = explode(",", $orderTags);
        $explodeTags = array_map('trim', $explodeTags);
        $explodeTags = array_map('strtolower', $explodeTags);
        if (in_array("extracted", $explodeTags)) {
            return true;
        }
        return false;
    }

    /**
     * Response function
     */
    public function returnData($status, $message, $error, $jsonCallback = null) {

        $returnData = array(
            'status' => $status,
            'message' => $message,
            'body' => $error

        );
        if (!empty($jsonCallback)) {

            echo $jsonCallback . '(' . json_encode($returnData) . ')';
            die();
        } else {

            return response()->json($returnData);
        }
    }
}
