<?php

namespace App\Jobs;

require_once(dirname(__FILE__).'/../../vendor/autoload.php');

use Illuminate\Bus\Queueable;
use Illuminate\Queue\SerializesModels;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use Illuminate\Support\Facades\File;
use Osiset\BasicShopifyAPI\BasicShopifyAPI;
use Osiset\BasicShopifyAPI\Options;
use Osiset\BasicShopifyAPI\Session;
use App\Models\IterableApi;
use Config;
use Log;
use DB;


class CreateShopifyCustomerJob implements ShouldQueue
{
    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    public $email;
    public $userName;
    public $partnerValue;
    public $session_id;
    public $shopify_api_version;
    public $iterableObj;
    
    /**
     * Create a new job instance.
     *
     * @return void
     */
    public function __construct($customerDetails) {

        $this->email = $customerDetails['email'];
        $this->userName = $customerDetails['userName'];
        $this->partnerValue = (isset($customerDetails['partnerValue']))?$customerDetails['partnerValue'] : '';
        $this->session_id = (isset($customerDetails['sessionId']))?$customerDetails['sessionId'] : '';
        $this->shopify_api_version = env('SHOPIFY_API_VERSION', true);
    }

    /**
     * Execute the job.
     * @return void
     */
    public function handle() {
        
        $this->iterableObj = new IterableApi();

        $password   = Config::get('shopify.passwordQuiz');
        $shopDomain = Config::get('shopify.shopDomainQuiz');

        $options = new Options();
        $options->setVersion($this->shopify_api_version);

        // Create the client and session
        $api = new BasicShopifyAPI($options);
        $api->setSession(new Session($shopDomain, $password));

        $randomPassword = $this->generateRandomString(8);
        
        $tags = [];
        if (trim($this->partnerValue) != '') {
            $tags[] = "partner:".$this->partnerValue;
        }
      
        $params = array(
            'customer' => array(
                'email' => $this->email,
                'first_name' => $this->userName,
                "send_email_welcome" => "false",
                "password" => $randomPassword,
                "password_confirmation" => $randomPassword,
                "tags" => $tags,
                //"accepts_marketing" => "true"
                "email_marketing_consent" => array(
                    "state" => "subscribed",
                    "opt_in_level" => "single_opt_in"
                )
            )
        );
        Log::channel('createShopifyCustomerLog')->info($this->partnerValue);

        Log::channel('createShopifyCustomerLog')->info(json_encode($params));
        $retry = false;
        $try = 1;
        do {
            Log::channel('createShopifyCustomerLog')->info($try.' try-create shopify customer ');
            $result = $api->rest('POST', '/admin/api/'.$this->shopify_api_version.'/customers.json', $params); 
            if (!isset($result['body']['customer']['id'])) {
                sleep(1);
                $retry = true;
                $try++;
            }
            if (isset($result['body']['customer']['id'])) {
                $this->updateRegistrationAuthInIterable();
            }
            
        } while($retry && $try <=3);
        Log::channel('createShopifyCustomerLog')->info("Customer create response:".json_encode($result));

        // if (empty($result['errors'])) {  //success

        //     // TO DO
        // } else {  //error

        //     // TO DO
        // }

    }
    
    /**
     * Update registration auth in iterable
     */
    public function updateRegistrationAuthInIterable() {

         if(!empty($this->email)){
            $registrationAuth = DB::table('users')->select('registration_auth')->where('user_email', '=', $this->email)->get();

            if(isset($registrationAuth[0]->registration_auth) && !empty($registrationAuth[0]->registration_auth)){
                $userIterableUpdateRequest = array(
                    "email" => $this->email,
                    "dataFields" => array(
                        "registrationAuth" => $registrationAuth[0]->registration_auth,
                        "sessionId" => $this->session_id
                    ),
                );
                $this->iterableObj->updateUserDataFields(json_encode($userIterableUpdateRequest));
            }

        }

    }
    
    /**
     * Generate random string
     */
    function generateRandomString($length) {

        $characters = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
        $charactersLength = strlen($characters);
        $randomString = '';
        for ($i = 0; $i < $length; $i++) {
            $randomString .= $characters[rand(0, $charactersLength - 1)];
        }
        return $randomString;
    }
}
