<?php

namespace App\Jobs;

use Illuminate\Bus\Queueable;
use Illuminate\Queue\SerializesModels;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use App\Models\Shopify\Order;
use App\Models\IterableApi;
use App\Models\RechargeApi;
use Config;
use Log;

class OrderStatusNotificationEventCreationJob implements ShouldQueue {

    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    public $webhookData;
    public $iterableObj;
    private $orderObj;
    public $rechargeObj;

    /**
     * Create a new job instance.
     */
    public function __construct($webhookDetails) {

        $this->webhookData = $webhookDetails;
    }

    /**
     * Execute the job.
     */
    public function handle() {

        $this->orderObj = new Order();
        $this->iterableObj = new IterableApi();
        $this->rechargeObj = new RechargeApi();
        $orderId = $this->webhookData->order_id;

        Log::channel('orderStatusNotificationLog')->info("Webhook processing started");

        // Get order details from Shopify
        $shopifyOrderDetails = $this->orderObj->getOrderDetailsById($orderId);
        if (!empty($shopifyOrderDetails['error']) || !empty($shopifyOrderDetails['errors']) ||
            empty($shopifyOrderDetails) || empty($shopifyOrderDetails['order'])) {

            $shopifyError  = "Unknown error";
            if (!empty($shopifyOrderDetails['error'])) {

                $shopifyError = json_encode($shopifyOrderDetails['error']);
            } else if (!empty($shopifyOrderDetails['errors'])) {

                $shopifyError = json_encode($shopifyOrderDetails['errors']);
            } else if (empty($shopifyOrderDetails) || empty($shopifyOrderDetails['order'])) {

                $shopifyError = 'Failed to get the order details from Shopify';
            }

            Log::channel('orderStatusNotificationLog')->info("Failed to get the order details 
            from Shopify(OrderId:".$orderId."), Error: (".$shopifyError.")");
            return null;
        }

        $orderGiftStatus = 0;
        $orderOtpStatus = 0;
        $orderShipmentStatus = '';
        $orderTags = !empty($shopifyOrderDetails['order']['tags']) ? 
        trim($shopifyOrderDetails['order']['tags']) : '';

        $shOrderTags = $this->orderTagFormatting($orderTags);

        if (in_array('gift', $shOrderTags) || in_array('prepaid', $shOrderTags)) {

            $orderGiftStatus = 1;
        }
        if (in_array('otp', $shOrderTags)) {

            $orderOtpStatus = 1;
        }
        $orderShipmentStatus = !empty($this->webhookData->shipment_status) ? 
        trim(strtolower($this->webhookData->shipment_status)) : 'null';

        // Get recharge order details by Shopify order Id
        $rechargeOrderDetails = $this->rechargeObj->getRechargeOrderByShopifyOrderId($orderId);
        if (!empty($rechargeOrderDetails->error) || !empty($rechargeOrderDetails->errors) ||
            empty($rechargeOrderDetails) || empty($rechargeOrderDetails->orders)) {

            $rechargeError  = "Unknown error";
            if (!empty($rechargeOrderDetails->error)) {

                $rechargeError = json_encode($rechargeOrderDetails->error);
            } else if (!empty($rechargeOrderDetails->errors)) {

                $rechargeError = json_encode($rechargeOrderDetails->errors);
            } else if (empty($rechargeOrderDetails) || empty($rechargeOrderDetails->orders)) {

                $rechargeError = 'Failed to get the order details from recharge';
            }

            $returnData = [
                'error' => $rechargeError
            ];

            Log::channel('orderStatusNotificationLog')->info("Failed to get the order details from 
            the recharge(ShopifyOrderId:".$orderId.",Error: (".$rechargeError.")");
        }

        $subscriptionAddressId = !empty($rechargeOrderDetails->orders[0]->address_id) ? 
        trim($rechargeOrderDetails->orders[0]->address_id) : '';

        $iterableParamArray = [];
        $orderVitamins = [];
        $gifteeFirstName = '';
        $gifteeLastName = '';
        $giftPurchaserFirstName = '';
        $giftPurchaserLastName = '';
        $giftPackName = '';
        $giftNotificationDate = '';
        $gifteeEmail = '';
        $giftChargeIntervalFrequency = '';
        
        if (!empty($this->webhookData->line_items[0]->properties) || 
        isset($this->webhookData->line_items[0]->properties)) {

            $vitaminTempArray = [];
            foreach ($this->webhookData->line_items[0]->properties as $propKey => $propValue) {

                $propertyName = '';
                $propertyValue = '';
                $explodePropertyValue = '';
                $propertyName = !empty($propValue->name) ? trim(strtolower($propValue->name)) : '';

                // Check and get the gift properties
                if ($orderGiftStatus == 1) {

                    switch ($propertyName) {

                        case (trim($propertyName) == 'giftee_first_name'):
                            $gifteeFirstName = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;

                        case (trim($propertyName) == 'giftee_last_name'):
                            $gifteeLastName = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;
    
                        case (trim($propertyName) == 'gift_senders_first_name'):
                            $giftPurchaserFirstName = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;
    
                        case (trim($propertyName) == 'gift_senders_last_name'):
                            $giftPurchaserLastName = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;
    
                        case (trim($propertyName) == 'gift_pack_name'):
                            $giftPackName = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;
    
                        case (trim($propertyName) == 'gift_notification_date'):
                            $giftNotificationDate = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;

                        case (trim($propertyName) == 'giftee_email'):
                            $gifteeEmail = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;

                        case (trim($propertyName) == 'gift_charge_interval_frequency'):
                            $giftChargeIntervalFrequency = !empty($propValue->value) ? 
                            trim($propValue->value) : '';
                            break;
                    }
                }

                if (strpos($propertyName, 'vitamin') !== false) {

                    $propertyValue = $propValue->value;
            
                    if (strpos($propertyValue, '|') !== false) {

                        $explodePropertyValue = explode('|', $propertyValue);

                        if (count($explodePropertyValue) == 5) {

                            $vitaminTempArray = [
                                "vitamin_name" => trim($explodePropertyValue[0]),
                                "quantity" => trim($explodePropertyValue[2]),
                                "variant" => trim($explodePropertyValue[1])
                            ];
                            $orderVitamins[] = $vitaminTempArray;
                            $vitaminTempArray = [];
                        }
                    }
                }
            }
        }

        // Checking the order event is already created or not by checking the tags
        $eventCreatedStatus = $this->checkOrderEventCreatedStatus($orderTags, $orderShipmentStatus, 
        $gifteeEmail);
        if (!empty($eventCreatedStatus['eventCreatedStatus']) && $eventCreatedStatus['eventCreatedStatus'] == 1) {

            Log::channel('orderStatusNotificationLog')->info("Skipped: The ".$eventCreatedStatus['type']." order ".
            $eventCreatedStatus['eventType']." event is already created in iterable, OrderId:". $orderId);
            Log::channel('orderStatusNotificationLog')->info("Webhook processing completed");
            return null;
        }

        $prepaidOrderShipmentNumber = '';
        if (!empty($subscriptionAddressId)) {

            // Get all order details from recharge by address ID
            $allOrderDetails = $this->rechargeObj->getRechargeOrdersByAddressId($subscriptionAddressId);
            if (!empty($allOrderDetails->error) || !empty($allOrderDetails->errors) ||
                empty($allOrderDetails) || empty($allOrderDetails->orders)) {

                $rechargeError  = "Unknown error";
                if (!empty($allOrderDetails->error)) {

                    $rechargeError = json_encode($allOrderDetails->error);
                } else if (!empty($allOrderDetails->errors)) {

                    $rechargeError = json_encode($allOrderDetails->errors);
                } else if (empty($allOrderDetails) || empty($allOrderDetails->orders)) {

                    $rechargeError = 'Failed to get the order details from recharge';
                }

                $returnData = [
                    'error' => $rechargeError
                ];

                Log::channel('orderStatusNotificationLog')->info("Failed to get the order details from 
                the recharge(orderId:".$orderId.", addressId:".$subscriptionAddressId."),
                Error: (".$rechargeError.")");
            }

            $successOrderCount = 0;
            $totalOrderCount = 0;
            foreach ($allOrderDetails->orders as $orderKey => $orderValue) {

                if (strtolower(trim($orderValue->status)) == 'success') {

                    $successOrderCount++;
                }
                $totalOrderCount++;
            }
            $prepaidOrderShipmentNumber = $successOrderCount.'/'.$totalOrderCount;
        }

        $iterableParamArray = [
            "order_number" => $shopifyOrderDetails['order']['name'],
            "email_address" => $shopifyOrderDetails['order']['email'],
            "tracking_url" => $this->webhookData->tracking_url,
            "tracking_number" => $this->webhookData->tracking_number,
            "tracking_company" => !empty($this->webhookData->tracking_company) ? 
            trim($this->webhookData->tracking_company) : '',
            "first_name" => $shopifyOrderDetails['order']['customer']['first_name'],
            "vitamins" => $orderVitamins,
            "tax_price" => $shopifyOrderDetails['order']['total_tax'],
            "subtotal_price" => $shopifyOrderDetails['order']['subtotal_price'],
            "shipping_price" => $shopifyOrderDetails['order']['shipping_lines'][0]['price'],
            "total_price" => $shopifyOrderDetails['order']['total_price'],
            "otp" => ($orderOtpStatus == 1) ? "true" : "false",
            "giftee_first_name" => $gifteeFirstName,
            "giftee_last_name" => $gifteeLastName,
            "gift_senders_first_name" => $giftPurchaserFirstName,
            "gift_senders_last_name" => $giftPurchaserLastName,
            "gift_pack_name" => $giftPackName,
            "gift_notification_date" => $giftNotificationDate,
            "giftee_email" => $gifteeEmail,
            "gift_charge_interval_frequency" => $giftChargeIntervalFrequency,
            "shipment_number" => $prepaidOrderShipmentNumber
        ];

        Log::channel('orderStatusNotificationLog')->info('IterableDataFields:'.json_encode($iterableParamArray));

        $eventCreatedResponse = '';
        if ($orderShipmentStatus == 'null') {

            // Shipped notification
            $eventCreatedResponse = $this->createShippedIterableEvent($iterableParamArray);
            if ($eventCreatedResponse['status'] == 'success') {
                $this->updateEventStatusToOrderTag($orderId, $orderTags, $eventCreatedResponse['eventName']);
            }

        } else if ($orderShipmentStatus == 'in_transit') {

            // Out for delivery notification
            $eventCreatedResponse = $this->createOutForDeliveryIterableEvent($iterableParamArray);
            if ($eventCreatedResponse['status'] == 'success') {
                $this->updateEventStatusToOrderTag($orderId, $orderTags, $eventCreatedResponse['eventName']);
            }

        } else if ($orderShipmentStatus == 'delivered') {

            // Delivered notification
            $eventCreatedResponse = $this->createDeliveredIterableEvent($iterableParamArray);
            if ($eventCreatedResponse['status'] == 'success') {
                $this->updateEventStatusToOrderTag($orderId, $orderTags, $eventCreatedResponse['eventName']);
            }
        }

        Log::channel('orderStatusNotificationLog')->info("Webhook processing completed");
    }

    /**
     * Check order iterable event created status
     */
    public function checkOrderEventCreatedStatus($orderTags, $orderShipmentStatus, $gifteeEmail) {

        $searchingRegularEvent = '';
        $searchingGiftEvent = '';
        $iterableEventCreatedStatus = 0;
        $orderType = '';
        $iterableEventToCreate = '';
        $resultData = [];

        switch ($orderShipmentStatus) {

            case (trim($orderShipmentStatus) == 'null'):
                $searchingRegularEvent = 'packShipped';
                $searchingGiftEvent = 'giftPackShipped';
                break;

            case (trim($orderShipmentStatus) == 'in_transit'):
                $searchingRegularEvent = 'packOutForDelivery';
                $searchingGiftEvent = 'giftPackOutForDelivery';
                break;

            case (trim($orderShipmentStatus) == 'delivered'):
                $searchingRegularEvent = 'packDelivered';
                $searchingGiftEvent = 'giftPackDelivered';
                break;
        }

        $shOrderTags = [];
        $shOrderTags = $this->orderTagFormatting($orderTags);

        if (!empty($gifteeEmail)) {

            if (in_array(strtolower($searchingGiftEvent), $shOrderTags)) {

                $iterableEventCreatedStatus = 1;
                $orderType = 'gift';
                $iterableEventToCreate = $searchingGiftEvent;
            }

        } else {

            if (in_array(strtolower($searchingRegularEvent), $shOrderTags)) {

                $iterableEventCreatedStatus = 1;
                $orderType = 'regular';
                $iterableEventToCreate = $searchingRegularEvent;
            }

        }

        $resultData = [
            "type" => $orderType,
            "eventType" => $iterableEventToCreate,
            "eventCreatedStatus" => $iterableEventCreatedStatus
        ];

        return $resultData;
    }

    /**
     * Order tag formatting
     */
    public function orderTagFormatting($orderTags) {

        $shOrderTags = [];
        if (!empty($orderTags)) {

            $shOrderTags = explode(",", $orderTags);
            $shOrderTags = array_map('trim', $shOrderTags);
            $shOrderTags = array_map('strtolower', $shOrderTags);
        }
        return $shOrderTags;
    }

    /**
     * Order tag update
     */
    public function updateEventStatusToOrderTag($shopifyOrderId, $orderTags, $newOrderTag) {

        $explodedOrderTags = [];
        if (!empty($orderTags)) {

            $explodedOrderTags = explode(",", $orderTags);
            $explodedOrderTags[] = $newOrderTag;
        } else {

            $explodedOrderTags[] = $newOrderTag;
        }
        $updatedShopifyTags = implode(",", $explodedOrderTags);

        $orderUpdateRequest = [];
        $orderUpdateRequest = array (
            "order" => array (
                "id" => $shopifyOrderId,
                "tags" => $updatedShopifyTags
            )
        );

        // Update Shopify order tags
        $shopifyOrderUpdateResponse = $this->orderObj->updateOrder($orderUpdateRequest, $shopifyOrderId);
        if (!empty($shopifyOrderUpdateResponse['error']) || !empty($shopifyOrderUpdateResponse['errors']) ||
            empty($shopifyOrderUpdateResponse)) {

            $shopifyError  = "Unknown error";
            if (!empty($shopifyOrderUpdateResponse['error'])) {

                $shopifyError = json_encode($shopifyOrderUpdateResponse['error']);
            } else if (!empty($shopifyOrderUpdateResponse['errors'])) {

                $shopifyError = json_encode($shopifyOrderUpdateResponse['errors']);
            } else if (empty($shopifyOrderUpdateResponse)) {

                $shopifyError = 'Failed to update the Shopify order';
            }
            $returnData = [
                'error' => $shopifyError
            ];

            Log::channel('orderStatusNotificationLog')->info("Failed to update the Shopify order
            (orderUpdateRequest:".json_encode($orderUpdateRequest)."), Error: (".$shopifyError.")");
        }
    }

    /**
     * Order shipped notification event creation
     */
    public function createShippedIterableEvent($iterableDataArray) {

        $shippedEventRequest = [];
        $shippedEventDataFieldArray = [];
        $responseData = [];
        $orderEmailAddress = '';
        $eventName = '';

        $orderEmailAddress = $iterableDataArray['email_address'];
        $eventName = 'packShipped';
        if (!empty($iterableDataArray['giftee_email'])) {

            $orderEmailAddress = $iterableDataArray['giftee_email'];
            $eventName = 'giftPackShipped';
        }

        $shippedEventDataFieldArray = $this->getIterableDataFields($iterableDataArray);

        $shippedEventRequest = array(
            "email" => strtolower($orderEmailAddress),
            "eventName" => $eventName,
            "dataFields" => $shippedEventDataFieldArray
        );

        $eventResponse = $this->iterableObj->createIterableEvent(json_encode($shippedEventRequest));
        if (isset($eventResponse->code) || trim(strtolower($eventResponse->code)) == 'success') {

            Log::channel('orderStatusNotificationLog')->info("Order shipped iterable event: - Successfully created the
            event for Email(".strtolower($orderEmailAddress).")");
            $responseData = [
                "status" => "success",
                "eventName" => $eventName
            ];
        } else {

            Log::channel('orderStatusNotificationLog')->info("Order shipped iterable event: - Failed to create the event 
            for Email (".strtolower($orderEmailAddress).")");
            $responseData = [
                "status" => "failed",
            ];
        }
        return $responseData;
    }

    /**
     * Order out for delivery notification event creation
     */
    public function createOutForDeliveryIterableEvent($iterableDataArray) {

        $shippedEventRequest = [];
        $shippedEventDataFieldArray = [];
        $responseData = [];
        $orderEmailAddress = '';
        $eventName = '';

        $orderEmailAddress = $iterableDataArray['email_address'];
        $eventName = 'packOutForDelivery';
        if (!empty($iterableDataArray['giftee_email'])) {

            $orderEmailAddress = $iterableDataArray['giftee_email'];
            $eventName = 'giftPackOutForDelivery';
        }

        $shippedEventDataFieldArray = $this->getIterableDataFields($iterableDataArray);

        $shippedEventRequest = array(
            "email" => strtolower($orderEmailAddress),
            "eventName" => $eventName,
            "dataFields" => $shippedEventDataFieldArray
        );

        $eventResponse = $this->iterableObj->createIterableEvent(json_encode($shippedEventRequest));
        if (isset($eventResponse->code) || trim(strtolower($eventResponse->code)) == 'success') {

            Log::channel('orderStatusNotificationLog')->info("Order out for delivery iterable event: - Successfully created the
            event for Email(".strtolower($orderEmailAddress).")");
            $responseData = [
                "status" => "success",
                "eventName" => $eventName
            ];
        } else {

            Log::channel('orderStatusNotificationLog')->info("Order out for delivery iterable event: - Failed to create the event 
            for Email (".strtolower($orderEmailAddress).")");
            $responseData = [
                "status" => "failed",
            ];
        }
        return $responseData;
    }

    /**
     * Order delivered notification event creation
     */
    public function createDeliveredIterableEvent($iterableDataArray) {

        $shippedEventRequest = [];
        $shippedEventDataFieldArray = [];
        $responseData = [];
        $orderEmailAddress = '';
        $eventName = '';

        $orderEmailAddress = $iterableDataArray['email_address'];
        $eventName = 'packDelivered';
        if (!empty($iterableDataArray['giftee_email'])) {

            $orderEmailAddress = $iterableDataArray['giftee_email'];
            $eventName = 'giftPackDelivered';
        }

        $shippedEventDataFieldArray = $this->getIterableDataFields($iterableDataArray);

        $shippedEventRequest = array(
            "email" => strtolower($orderEmailAddress),
            "eventName" => $eventName,
            "dataFields" => $shippedEventDataFieldArray
        );

        $eventResponse = $this->iterableObj->createIterableEvent(json_encode($shippedEventRequest));
        if (isset($eventResponse->code) || trim(strtolower($eventResponse->code)) == 'success') {

            Log::channel('orderStatusNotificationLog')->info("Order delivered iterable event: - Successfully created the
            event for Email(".strtolower($orderEmailAddress).")");
            $responseData = [
                "status" => "success",
                "eventName" => $eventName
            ];
        } else {

            Log::channel('orderStatusNotificationLog')->info("Order delivered iterable event: - Failed to create the event 
            for Email (".strtolower($orderEmailAddress).")");
            $responseData = [
                "status" => "failed",
            ];
        }
        return $responseData;
    }

    /**
     * Iterable event data field parameter handling
     */
    public function getIterableDataFields($iterableDataArray) {

        $shippedEventDataFieldArray = [];

        if (!empty($iterableDataArray['giftee_email'])) {
            $shippedEventDataFieldArray = [
                "order_number" => $iterableDataArray['order_number'],
                "tracking_url" => $iterableDataArray['tracking_url'],
                "tracking_number" => $iterableDataArray['tracking_number'],
                "tracking_company" => $iterableDataArray['tracking_company'],
                "vitamins" => $iterableDataArray['vitamins'],
                "tax_price" => $iterableDataArray['tax_price'],
                "subtotal_price" => $iterableDataArray['subtotal_price'],
                "shipping_price" => $iterableDataArray['shipping_price'],
                "total_price" => $iterableDataArray['total_price'],
                "giftee_first_name" => $iterableDataArray['giftee_first_name'],
                "giftee_last_name" => $iterableDataArray['giftee_last_name'],
                "gift_senders_first_name" => $iterableDataArray['gift_senders_first_name'],
                "gift_senders_last_name" => $iterableDataArray['gift_senders_last_name'],
                "gift_pack_name" => $iterableDataArray['gift_pack_name'],
                "gift_notification_date" => $iterableDataArray['gift_notification_date'],
                "giftee_email" => $iterableDataArray['giftee_email'],
                "gift_charge_interval_frequency" => $iterableDataArray['gift_charge_interval_frequency'],
                "shipment_number" => $iterableDataArray['shipment_number']
            ];
        } else {

            $shippedEventDataFieldArray = [
                "order_number" => $iterableDataArray['order_number'],
                "tracking_url" => $iterableDataArray['tracking_url'],
                "tracking_number" => $iterableDataArray['tracking_number'],
                "tracking_company" => $iterableDataArray['tracking_company'],
                "first_name" => $iterableDataArray['first_name'],
                "vitamins" => $iterableDataArray['vitamins'],
                "tax_price" => $iterableDataArray['tax_price'],
                "subtotal_price" => $iterableDataArray['subtotal_price'],
                "shipping_price" => $iterableDataArray['shipping_price'],
                "total_price" => $iterableDataArray['total_price'],
                "otp" => $iterableDataArray['otp']
            ];
        }
        return $shippedEventDataFieldArray;
    }

}
