<?php

namespace App\Jobs;

use Illuminate\Bus\Queueable;
use Illuminate\Queue\SerializesModels;
use Illuminate\Queue\InteractsWithQueue;
use Illuminate\Contracts\Queue\ShouldQueue;
use Illuminate\Foundation\Bus\Dispatchable;
use Osiset\BasicShopifyAPI\BasicShopifyAPI;
use Osiset\BasicShopifyAPI\Options;
use Osiset\BasicShopifyAPI\Session;
use App\Models\Shopify\Customer;
use Config;
use DB;
use Log;


class TagUserJob implements ShouldQueue
{
    use Dispatchable, InteractsWithQueue, Queueable, SerializesModels;

    public $recommendation;
    public $tagQuestion;
    public $sessionID;
    public $shopify_api_version;
    public $api;
    public $customerObj;

    /**
     * Create a new job instance.
     *
     * @return void
     */
    public function __construct($recommendation, $tagQuestion, $sessionID) {

        $this->recommendation = $recommendation;
        $this->tagQuestion = $tagQuestion;
        $this->sessionID = $sessionID;
        $this->shopify_api_version = env('SHOPIFY_API_VERSION', true);
    }

    /**
     * Execute the job.
     *
     * @return void
     */
    public function handle() {

        $this->customerObj = new Customer();
        $recommendation = json_decode($this->recommendation);
        if (isset($recommendation->Result)) {

            $recommendation = $recommendation->Result;
        }

        $newTags = $this->getRecommendationTags($recommendation);

        $password   = Config::get('shopify.passwordQuiz');
        $shopDomain = Config::get('shopify.shopDomainQuiz');

        $options = new Options();
        $options->setVersion($this->shopify_api_version);

        // Create the client and session
        $this->api = new BasicShopifyAPI($options);
        $this->api->setSession(new Session($shopDomain, $password));

        $quizUser = DB::table('quiz_user')->where('session_id', '=', $this->sessionID)->get();

        if(isset($quizUser[0]->user_id))
        {
            $userId = $quizUser[0]->user_id;

            $userEmail = DB::table('users')->where('id', '=', $userId)->get();
            $emailId = !empty($userEmail[0]->user_email) ? trim($userEmail[0]->user_email) : "";

            if (empty($emailId)) {
                Log::channel('quizControllerLog')->info("TagUserJob: Not found the email address in DB(sessionId:".$this->sessionID.")");
                return false;
            }
    
            $customerDetails = $this->customerObj->getCustomerDetailBySearch($emailId);

            if (!empty($customerDetails['errors']) || !empty($customerDetails['error'])) {

                $error = "Unknown error";
                if (!empty($customerDetails['error'])) {

                    $error = json_encode($customerDetails['error']);
                } elseif (!empty($customerDetails['errors'])) {
                    
                    $error = json_encode($customerDetails['errors']);
                }
                if (empty($error)) {

                    $error = "An error occurred while getting the customer details from Shopify (EmailID - " . $emailId . ")";
                }
                $returnData = [
                    "error" => $error
                ];

                Log::channel('quizControllerLog')->info("TagUserJob: Failed to get the customer details from Shopify(Email:".$emailId."), 
                Error: (".$error.")");
            }
            
            if (!empty($customerDetails['customers'][0]) && isset($customerDetails['customers'][0])) {

                $customer = [];
                foreach ($customerDetails['customers'] as $cust) {
                    if (trim(strtolower($cust['email'])) == trim(strtolower($emailId))) {
                        $customer = $cust;
                        break;
                    }
                }

                if (empty($customer)) {
                    Log::channel('quizControllerLog')->info("TagUserJob: Not found the searched email address in Shopify(Email: ".$emailId.")");
                    return false;
                }

                $customerId = $customer['id'];
                $tags = $customer['tags'];

                $newTags = $this->setNewtags($tags, $newTags);

                $params = array(
                    'customer' => array(
                        'id' => $customerId,
                        'tags' => $newTags   
                    )
                );
                
                $try = 1;
                $retry = false;
                do {
                    $result = $this->api->rest('PUT', '/admin/api/'.$this->shopify_api_version.'/customers/'.$customerId.'.json', $params); 
                    if (!isset($result['body']['customer']['id'])) {
                        sleep(1);
                        $retry = true;
                        $try++;
                    }
                } while($retry && $try <=3);
            }     
        }
    }

    /**
     * Set recommendation tags
     */
    function getRecommendationTags($recommendation) {

        $allergens    = $recommendation->Allergens;
        $deficiencies = $recommendation->Deficiencies;
        $recDetails   = $recommendation->Recommendation;
        $sessionID    = $this->sessionID;

        $tag = '';
        $allergensTag = '';
        $deficiencyTag = '';
        $recommendationTag = '';

        foreach($allergens as $key => $data) {

            if ($data == true) {

                $allergensTag.= "myAllergens:".$key.",";
            }
        }

        foreach ($deficiencies as $key => $data) {

            if ($data == true) {

                $deficiencyTag.= "myDeficiencies:".$key.",";
            }
        }

        foreach ($recDetails as $value) {

            $name = $value->Name;
            if ($name == 'Hair, Skin & Nails') {
                $name = 'Hair|Skin|Nails';
            }
            $recommendationTag.= " myRecommendations:".$name.",";
        }

        $tag .= "Gender:".$recommendation->SexProfile.",";

        if ($recommendation->North37 == '1') {
            $tag .= 'North37,';
        }

        if ($recommendation->ProstateInterest == '1') {
            $tag .= 'ProstateInterest,';
        }

        if ($recommendation->PrenatalInterest == '1') {
            $tag .= 'PrenatalInterest,';
        }

        if ($recommendation->PostnatalInterest == '1') {
            $tag .= 'PostnatalInterest,';
        }

        if ($recommendation->HealthyMentrualCycleInterest == '1') {
            $tag .= 'HealthyMentrualCycleInterest,';
        }

        if ($recommendation->PerimenopauseInterest == '1') {
            $tag .= 'PerimenopauseInterest,';
        }

        if ($recommendation->MenopauseInterest == '1') {
            $tag .= 'MenopauseInterest,';
        }

        if ($recommendation->PlanningPregnancy == '1') {
            $tag .= 'PlanningPregnancy,';
        }

        if ($recommendation->TakingMedication == '1') {
            $tag .= 'TakingMedication,';
        }

        if ($recommendation->BloodClottingIssue == '1') {
            $tag .= 'BloodClottingIssue,';
        }
        
        if ($recommendation->Pregnant == '1') {
            $tag .= 'Pregnant,';
        }

        if ($recommendation->Breastfeeding == '1') {
            $tag .= 'Breastfeeding,';
        }

        $tag .= $allergensTag;
        $tag .= $deficiencyTag;
        $tag .= $recommendationTag;

        if (!empty($this->tagQuestion)) {

            foreach($this->tagQuestion[14] as $data) {

                if ($data == 'Hair, Skin or Nails') {
                    $data = 'Hair|Skin|Nails';
                }
                $tag .= "myHealthInterests:".$data.",";
            }

            foreach($this->tagQuestion[32] as $data) {

                $tag .= "myInterests-Activities:".$data.",";
            }
        }
        

        $tag .= "QuizCompleted";

        return $tag;
    }

    /**
     * Set new tags
     */
    function setNewtags($tags, $recommendationTags) {

        $tagExploded = explode(',', $tags);

        $tagCollection = array('myAllergens', 'myDeficiencies', 'myRecommendations',
            'Gender', 'North37', 'ProstateInterest', 'PrenatalInterest', 'PostnatalInterest',
            'HealthyMentrualCycleInterest', 'PerimenopauseInterest', 'MenopauseInterest',
            'PlanningPregnancy', 'TakingMedication', 'BloodClottingIssue', 'Pregnant',
            'Breastfeeding', 'myHealthInterests', 'myInterests-Activities'
        );

        if (!empty($tagExploded)) {

            foreach($tagExploded as $key => $tag) {

                foreach($tagCollection as $recTag) {

                    if (strpos(trim($tag), $recTag) !== FALSE) {
                        unset($tagExploded[$key]);
                    } else {

                        continue;
                    }
                }
            }
        }

        $newTags = implode(",", $tagExploded).",".$recommendationTags;
        return $newTags;
    }
}
