<?php

namespace App\Model\Shopify;

use OhMyBrew\BasicShopifyAPI;
use Exception;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\ClientException;
use GuzzleHttp\Exception\ServerException;
use GuzzleHttp\Psr7\Request;
use GuzzleHttp\Psr7\Uri;

class ExtentedShopifyAPI extends BasicShopifyAPI
{

    /**
    * Used for handling the API rate limit
    */
    protected $retryCall;

    public function __construct()
    {
        parent::__construct();
        // Set the retryCall as 1 for handling the API rate limit
        $this->retryCall = 1;

        $this->enableRateLimiting();
    }

    /**
     * Runs a request to the Shopify API.
     *
     * @param string     $type   The type of request... GET, POST, PUT, DELETE
     * @param string     $path   The Shopify API path... /admin/xxxx/xxxx.json
     * @param array|null $params Optional parameters to send with the request
     *
     * @throws Exception
     *
     * @return object An Object of the Guzzle response, and JSON-decoded body
     */
    public function rest(string $type, string $path, array $params = null)
    {
       
        // Check the rate limit before firing the request
        if ($this->isRateLimitingEnabled() && $this->requestTimestamp) {
            // Calculate in milliseconds the duration the API call took
            $duration = round(microtime(true) - $this->requestTimestamp, 3) * 1000;
            $waitTime = ($this->rateLimitCycle - $duration) + $this->rateLimitCycleBuffer;

            if ($waitTime > 0) {
                // Do the sleep for X mircoseconds (convert from milliseconds)
                $this->log('Rest rate limit hit');
                usleep($waitTime * 1000);
            }
        }

        // Update the timestamp of the request
        $tmpTimestamp = $this->requestTimestamp;
        $this->requestTimestamp = microtime(true);

        $errors = false;
        $response = null;
        $body = null;

        try {
            // Build URI and try the request
            $uri = $this->getBaseUri()->withPath($this->versionPath($path));

            // Build the request parameters for Guzzle
            $guzzleParams = [];
            if ($params !== null) {
                $guzzleParams[strtoupper($type) === 'GET' ? 'query' : 'json'] = $params;
            }

            $this->log("[{$uri}:{$type}] Request Params: ".json_encode($params));

            // Set the response
            $response = $this->client->request($type, $uri, $guzzleParams);
            $body = $response->getBody();
        } catch (Exception $e) {
            if ($e instanceof ClientException || $e instanceof ServerException) {
                // 400 or 500 level error, set the response
                $response = $e->getResponse();
                $body = $response->getBody();

                // Build the error object
                $errors = (object) [
                    'status'    => $response->getStatusCode(),
                    'body'      => $this->jsonDecode($body),
                    'exception' => $e,
                ];

                $this->log("[{$uri}:{$type}] {$response->getStatusCode()} Error: {$body}");
            } else {
                // Else, rethrow
                throw $e;
            }
        }

        // Grab the API call limit header returned from Shopify
        $callLimitHeader = $response->getHeader('http_x_shopify_shop_api_call_limit');
        if ($callLimitHeader) {
            $calls = explode('/', $callLimitHeader[0]);
            $this->apiCallLimits['rest'] = [
                'left'  => (int) $calls[1] - $calls[0],
                'made'  => (int) $calls[0],
                'limit' => (int) $calls[1],
            ];
        }

        $this->log("[{$uri}:{$type}] {$response->getStatusCode()}: ".json_encode($errors ? $body->getContents() : $body));

        if (!empty($errors) && empty($body->getContents())) {
            
            if ($errors->status == '429') {
                Log::channel('ohMyBrew')->info(json_encode(debug_backtrace()));
                Log::channel('ohMyBrew')->info("Api rate limit exceeds");
                if ($this->retryCall <=3) {
                    $this->retryCall++;
                    Log::channel('ohMyBrew')->info("Retrying the API due to rate limit exceeds");
                    return $this->rest($type, $path, $params);
                }
            }
        }
        // Return Guzzle response and JSON-decoded body
        return (object) [
            'response'   => $response,
            'errors'     => $errors,
            'body'       => $errors ? $body->getContents() : $this->jsonDecode($body),
            'timestamps' => [$tmpTimestamp, $this->requestTimestamp],
        ];
    }
}
